#!/usr/bin/env python

import sys

# encapsulate file handling and writing 
class fileManager(object):
    def __init__(self, filename):
        self.file = open(filename,'wb')
        self.BUFFERSIZE = 256
        self.bufferlen = 0
        self.elemsize = 32
        self.buffer = self.BUFFERSIZE*[0xFFFFFFFF]
        for i in range(0,self.BUFFERSIZE):
            self.buffer[i] = 0
            
    def writeToFile(self, bits, len):
        idx = self.bufferlen / self.elemsize;
        pos = self.bufferlen % self.elemsize;
        self.buffer[idx] |= (bits << pos);
        maxwritten = self.elemsize-pos;
        notwritten = len - maxwritten;
    
        if notwritten > 0:
            self.bufferlen += maxwritten;
        else:
            self.bufferlen += len;
      
        if self.bufferlen == self.elemsize*self.BUFFERSIZE: #buffer full => write buffer to file
            for i in range(0, BUFFERSIZE):
                self.file.write(chr(0xFF & self.buffer[i]))
                self.file.write(chr(0xFF & (self.buffer[i] >> 8)))
                self.file.write(chr(0xFF & (self.buffer[i] >> 16)))
                self.file.write(chr(0xFF & (self.buffer[i] >> 24)))
            self.bufferlen = 0;
            for i in range (0,self.BUFFERSIZE):
                self.buffer[i] = 0;
    
        idx = self.bufferlen / self.elemsize;
        if notwritten > 0: #not yet all bits written
            self.buffer[idx] |= (bits >> (maxwritten));
            self.bufferlen += notwritten;
    
    def writeBuffer(self):
        self.bufferlen -= 1;
        idx = self.bufferlen / self.elemsize;
        pos = self.bufferlen % self.elemsize;
        if pos > 0:
            idx += 1;
        if self.bufferlen >= 0: #not all bits from buffer written
            for i in range(0, idx):
                self.file.write(chr(0xFF & self.buffer[i]))
                self.file.write(chr(0xFF & (self.buffer[i] >> 8)))
                self.file.write(chr(0xFF & (self.buffer[i] >> 16)))
                self.file.write(chr(0xFF & (self.buffer[i] >> 24)))
        self.bufferlen = -1;


    def finish(self):
        self.writeBuffer()
        self.file.close()

# compresses an arbitrary file        
class compress(object):
    # infilename is the filename of the file you want to compress
    # outfilename is the filename of the compressed file
    def __init__(self, infilename, outfilename):
        # open file for reading
        file = open(infilename,'r')
        # read file into a list
        self.text = file.read()
        # close file
        file.close()
        self.outfilename = outfilename
        # generate list of structure [[,],....[,]]
        self.huffcode = []
        for i in range(0,256):
            self.huffcode.append([0,0]) # first entry "= length of code in Bits, second entry := code stored in an Integer variable
    
    
    def calcHuffmanCode(self, symcount):
        # ----------TODO begin---------------
        #replace the above code with your implementation of huffman-code calculation
        
        for i in range(0,len(symcount)):
            self.huffcode[symcount[i][1]][0] = 8
            self.huffcode[symcount[i][1]][1] = symcount[i][1]
        # ----------TODO end---------------
    
    
    def analyzeText(self):
        #initialize  counter
        symbol_count = []
        for i in range(0,256):
            symbol_count.append([0,i])
        
        #calculate occurrence for letters
        for i in range(0,len(self.text)):
            symbol_count[ord(self.text[i])][0] += 1
            
        # sort letters after occurrence
        symbol_count.sort(lambda x,y: y[0]-x[0])
        symnonzero = 255
        # delete all letters which don't occure in text
        while symbol_count[symnonzero][0] == 0:
            symnonzero -= 1
        symnonzero += 1
        symbol_count = symbol_count[:symnonzero]
        
        # calculate relative occurrence
        for i in range(0,symnonzero):
            symbol_count[i][0] /= len(self.text)
            
        # calculate huffman-code for all letters
        self.calcHuffmanCode(symbol_count)

        
    def doCompression(self):
        # read inputfile for compression
        self.analyzeText()
        # prepare to write Bitwise into compressed file
        f = fileManager(self.outfilename)
        # write number of letters/length of inputfile to compressed file
        f.writeToFile(len(self.text),32)
        # for each letter, read the huffman-code and write it to compressed file
        for i in range(0,len(self.text)):
            f.writeToFile(self.huffcode[ord(self.text[i])][1],self.huffcode[ord(self.text[i])][0])
        # some cleanups ( clear buffer, close file)
        f.finish()


if len(sys.argv) < 3:
    print 'Wrong argumentlist!'
    print 'correct usage: huffman_encoding.py inputfile outputfile'
else:
    c = compress(sys.argv[1],sys.argv[2])
    c.doCompression()
